==> zaailabcorelib-0.2.0.6-py3-none-any.whl: zaailabcorelib/zserver/thrift_server/TMultiPoolServer.py <==

import time
from queue import Empty
from multiprocessing import Queue
import multiprocessing
import logging
import select
import socket
import struct
import threading
from collections import deque
from six.moves import queue
from zaailabcorelib.thrift.transport import TTransport
from zaailabcorelib.thrift.protocol.TBinaryProtocol import TBinaryProtocolFactory
logger = logging.getLogger(__name__)

__all__ = ['TMultiPoolServer', 'TModelBase', 'THandlerBase']


def current_milli_time(): return int(round(time.time() * 1000))


def current_nano_time(): return int(round(time.time() * 1000000))


class TModelBase(multiprocessing.Process):
    def __init__(self, inference_queue, result_dict, model_config, batch_infer_size=1, batch_timeout=10):
        super(TModelBase, self).__init__()
        self.inference_queue = inference_queue
        self.result_dict = result_dict
        self.batch_infer_size = batch_infer_size
        self.batch_timeout_in_sec = self._microsec_to_sec(batch_timeout)
        self.batch_timeout = batch_timeout
        self.model = self.model_init(model_config)

    def _microsec_to_sec(self, microsec):
        return microsec/1000

    def model_init(self, model_config: dict):
        raise NotImplementedError

    def predict(self, list_input):
        raise NotImplementedError

    # def run(self):
    #     list_inference = []
    #     list_request_id = []
    #     while True:
    #         t = current_milli_time()
    #         while (len(list_inference) < self.batch_infer_size) and current_milli_time() - t < self.batch_timeout:
    #             try:
    #                 [request_id, inp] = self.inference_queue.get(block=True, timeout=self.batch_timeout_in_sec)
    #                 list_inference.append(inp)
    #                 list_request_id.append(request_id)
    #             except Empty:
    #                 pass
    #             t = current_milli_time()
    #         list_result = self.predict(list_inference)
    #         for [_id, res] in zip(list_request_id, list_result):
    #             self.result_dict.update({_id:res})
    #         list_request_id.clear()
    #         list_inference.clear()

    def run(self):
        list_inference = []
        list_request_id = []
        while True:
            while True:
                try:
                    [request_id, inp] = self.inference_queue.get(block=False) #, timeout=self.batch_timeout_in_sec)
                    list_inference.append(inp)
                    list_request_id.append(request_id)
                    if (len(list_inference) < self.batch_infer_size):
                        continue
                except Empty:
                    break
            if len(list_inference) != 0:
                list_result = self.predict(list_inference)
                for [_id, res] in zip(list_request_id, list_result):
                    self.result_dict.update({_id:res})
                list_request_id.clear()
                list_inference.clear()

class THandlerBase():
    def __init__(self, inference_queue, result_dict):
        self.inference_queue = inference_queue
        self.result_dict = result_dict

    def send_to_model(self, input_val):
        request_id = current_nano_time()
        input_ = self.process_input(input_val)
        self.inference_queue.put([request_id, input_], block=False)
        done = False
        while not done:
            result = self.result_dict.get(request_id, None)
            if result is not None:
                done = True
        return result

    def process_input(self, input):
        raise NotImplementedError


class ConnectionSink(threading.Thread):
    """Worker is a small helper to process incoming connection."""

    def __init__(self, thread_id, queue, processor_cls, handler_cls, inference_queue, result_dict):
        threading.Thread.__init__(self)
        self.thread_id = thread_id
        self.queue = queue
        self.processor_cls = processor_cls
        self.handler_cls = handler_cls
        self.inference_queue = inference_queue
        self.result_dict = result_dict

    def run(self):
        """Process queries from task queue, stop if processor is None."""
        self.handler = self.handler_cls(
            inference_queue=self.inference_queue, result_dict=self.result_dict)
        self.processor = self.processor_cls(self.handler)
        while True:
            try:
                iprot, oprot, otrans, callback = self.queue.get()
                # if iprot is None:
                #     continue
                self.processor.process(iprot, oprot)
                callback(True, otrans.getvalue())
            except Exception:
                logger.exception(
                    "Exception while processing request")
                callback(False, b'')


WAIT_LEN = 0
WAIT_MESSAGE = 1
WAIT_PROCESS = 2
SEND_ANSWER = 3
CLOSED = 4


def locked(func):
    """Decorator which locks self.lock."""

    def nested(self, *args, **kwargs):
        self.lock.acquire()
        try:
            return func(self, *args, **kwargs)
        finally:
            self.lock.release()
    return nested


def socket_exception(func):
    """Decorator close object on socket.error."""

    def read(self, *args, **kwargs):
        try:
            return func(self, *args, **kwargs)
        except socket.error:
            logger.debug('ignoring socket exception', exc_info=True)
            self.close()
    return read


class Message(object):
    def __init__(self, offset, len_, header):
        self.offset = offset
        self.len = len_
        self.buffer = None
        self.is_header = header

    @property
    def end(self):
        return self.offset + self.len


class Connection(object):
    """Basic class is represented connection.

    It can be in state:
        WAIT_LEN --- connection is reading request len.
        WAIT_MESSAGE --- connection is reading request.
        WAIT_PROCESS --- connection has just read whole request and
                         waits for call ready routine.
        SEND_ANSWER --- connection is sending answer string (including length
                        of answer).
        CLOSED --- socket was closed and connection should be deleted.
    """

    def __init__(self, new_socket, wake_up):
        self.socket = new_socket
        self.socket.setblocking(False)
        self.status = WAIT_LEN
        self.len = 0
        self.received = deque()
        self._reading = Message(0, 4, True)
        self._rbuf = b''
        self._wbuf = b''
        self.lock = threading.Lock()
        self.wake_up = wake_up
        self.remaining = False

    @socket_exception
    def read(self):
        """Reads data from stream and switch state."""
        assert self.status in (WAIT_LEN, WAIT_MESSAGE)
        assert not self.received
        buf_size = 8192
        first = True
        done = False
        while not done:
            read = self.socket.recv(buf_size)
            rlen = len(read)
            done = rlen < buf_size
            self._rbuf += read
            if first and rlen == 0:
                if self.status != WAIT_LEN or self._rbuf:
                    logger.error('could not read frame from socket')
                else:
                    logger.debug(
                        'read zero length. client might have disconnected')
                self.close()
            while len(self._rbuf) >= self._reading.end:
                if self._reading.is_header:
                    mlen, = struct.unpack('!i', self._rbuf[:4])
                    self._reading = Message(self._reading.end, mlen, False)
                    self.status = WAIT_MESSAGE
                else:
                    self._reading.buffer = self._rbuf
                    self.received.append(self._reading)
                    self._rbuf = self._rbuf[self._reading.end:]
                    self._reading = Message(0, 4, True)
                    done = True
            first = False
            if self.received:
                self.status = WAIT_PROCESS
                break
        self.remaining = not done

    @socket_exception
    def write(self):
        """Writes data from socket and switch state."""
        assert self.status == SEND_ANSWER
        sent = self.socket.send(self._wbuf)
        if sent == len(self._wbuf):
            self.status = WAIT_LEN
            self._wbuf = b''
            self.len = 0
        else:
            self._wbuf = self._wbuf[sent:]
            

    @locked
    def ready(self, all_ok, message):
        """Callback function for switching state and waking up main thread.

        This function is the only function witch can be called asynchronous.

        The ready can switch Connection to three states:
            WAIT_LEN if request was oneway.
            SEND_ANSWER if request was processed in normal way.
            CLOSED if request throws unexpected exception.

        The one wakes up main thread.
        """
        assert self.status == WAIT_PROCESS
        if not all_ok:
            self.close()
            self.wake_up()
            return
        self.len = 0
        if len(message) == 0:
            # it was a oneway request, do not write answer
            self._wbuf = b''
            self.status = WAIT_LEN
        else:
            self._wbuf = struct.pack('!i', len(message)) + message
            self.status = SEND_ANSWER
        self.wake_up()

    @locked
    def is_writeable(self):
        """Return True if connection should be added to write list of select"""
        return self.status == SEND_ANSWER

    # it's not necessary, but...
    @locked
    def is_readable(self):
        """Return True if connection should be added to read list of select"""
        return self.status in (WAIT_LEN, WAIT_MESSAGE)

    @locked
    def is_closed(self):
        """Returns True if connection is closed."""
        return self.status == CLOSED

    def fileno(self):
        """Returns the file descriptor of the associated socket."""
        return self.socket.fileno()

    def close(self):
        """Closes connection"""
        self.status = CLOSED
        self.socket.close()


class TMultiPoolServer(object):
    """TMultiPoolServer."""

    def __init__(self,
                 handler_cls,
                 processor_cls,
                 model_cls,
                 lsocket,
                 list_model_config,
                 inputProtocolFactory=None,
                 outputProtocolFactory=None,
                 batch_infer_size=1,
                 batch_timeout=10,
                 n_models=1,
                 n_handlers=10):
        self.model_cls = model_cls
        self.handler_cls = handler_cls
        self.processor_cls = processor_cls
        self.socket = lsocket
        self.in_protocol = inputProtocolFactory or TBinaryProtocolFactory()
        self.out_protocol = outputProtocolFactory or self.in_protocol
        self.n_handlers = int(n_handlers)
        self.n_models = int(n_models)
        self.clients = {}
        self.connection_queue = queue.Queue()
        # self.inference_queue = queue.Queue()
        self.inference_queue = multiprocessing.Queue()
        self.result_dict = multiprocessing.Manager().dict()
        # self.result_dict = dict()
        self.list_model_config = list_model_config
        self.batch_infer_size = batch_infer_size
        self.batch_timeout = batch_timeout

        self._read, self._write = socket.socketpair()
        self.prepared = False
        self._stop = False
        self.list_handlers = []
        self.list_models = []

    def set_num_handler(self, num):
        """Set the number of worker threads that should be created."""
        assert not self.prepared, "Can't change number of threads after start"
        self.n_handlers = num

    def prepare(self):
        """Prepares server for serve requests."""
        if self.prepared:
            return
        self.socket.listen()

        for model_config in self.list_model_config:
            process = self.model_cls(inference_queue=self.inference_queue,
                                     result_dict=self.result_dict,
                                     model_config=model_config,
                                     batch_infer_size=self.batch_infer_size,
                                     batch_timeout=self.batch_timeout)
            process.start()
            self.list_models.append(process)

        for idx in range(self.n_handlers):
            thread = ConnectionSink(idx, self.connection_queue,
                                    self.processor_cls, self.handler_cls,
                                    self.inference_queue, self.result_dict)
            thread.setDaemon(True)
            thread.start()
            self.list_handlers.append(thread)
        self.prepared = True

    def wake_up(self):
        """Wake up main thread.

        The server usually waits in select call in we should terminate one.
        The simplest way is using socketpair.

        Select always wait to read from the first socket of socketpair.

        In this case, we can just write anything to the second socket from
        socketpair.
        """
        self._write.send(b'1')

    def stop(self):
        """Stop the server.

        This method causes the serve() method to return.  stop() may be invoked
        from within your handler, or from another thread.

        After stop() is called, serve() will return but the server will still
        be listening on the socket.  serve() may then be called again to resume
        processing requests.  Alternatively, close() may be called after
        serve() returns to close the server socket and shutdown all worker
        threads.
        """
        self._stop = True
        self.wake_up()

    def _select(self):
        """Does select on open connections."""
        readable = [self.socket.handle.fileno(), self._read.fileno()]
        writable = []
        remaining = []
        for i, connection in list(self.clients.items()):
            if connection.is_readable():
                readable.append(connection.fileno())
                if connection.remaining or connection.received:
                    remaining.append(connection.fileno())
            if connection.is_writeable():
                writable.append(connection.fileno())
            if connection.is_closed():
                del self.clients[i]
        if remaining:
            return remaining, [], [], False
        else:
            return select.select(readable, writable, readable) + (True,)

    def handle(self):
        """Handle requests.

        WARNING! You must call prepare() BEFORE calling handle()
        """
        assert self.prepared, "You have to call prepare before handle"
        rset, wset, xset, selected = self._select()
        for readable in rset:
            if readable == self._read.fileno():
                # don't care i just need to clean readable flag
                self._read.recv(1024)
            elif readable == self.socket.handle.fileno():
                try:
                    client = self.socket.accept()
                    if client:
                        self.clients[client.handle.fileno()] = Connection(client.handle,
                                                                          self.wake_up)
                except socket.error:
                    logger.debug('error while accepting', exc_info=True)
            else:
                connection = self.clients[readable]
                if selected:
                    connection.read()
                if connection.received:
                    connection.status = WAIT_PROCESS
                    msg = connection.received.popleft()
                    itransport = TTransport.TMemoryBuffer(
                        msg.buffer, msg.offset)
                    otransport = TTransport.TMemoryBuffer()
                    iprot = self.in_protocol.getProtocol(itransport)
                    oprot = self.out_protocol.getProtocol(otransport)
                    self.connection_queue.put([iprot, oprot,
                                               otransport, connection.ready])
        for writeable in wset:
            self.clients[writeable].write()
        for oob in xset:
            self.clients[oob].close()
            del self.clients[oob]

    def close(self):
        """Closes the server."""
        for _ in range(self.n_handlers):
            self.connection_queue.put([None, None, None, None, None])
        self.socket.close()
        self.prepared = False

    def serve(self):
        """Serve requests.

        Serve requests forever, or until stop() is called.
        """
        self._stop = False
        self.prepare()
        while not self._stop:
            self.handle()

==> zaailabcorelib-0.2.1.8-py3-none-any.whl: zaailabcorelib/zserver/thrift_server/TMultiPoolServer.py <==

import time
from queue import Empty
from multiprocessing import Queue
import multiprocessing
import logging
import select
import socket
import struct
import threading
from collections import deque
from six.moves import queue
from zaailabcorelib.thrift.transport import TTransport
from zaailabcorelib.thrift.protocol.TBinaryProtocol import TBinaryProtocolFactory
from .utils import current_milli_time, current_nano_time
logger = logging.getLogger(__name__)

__all__ = ['TMultiPoolServer', 'TModelBase', 'THandlerBase']


class TModelBase(multiprocessing.Process):
    def __init__(self, inference_queue, result_dict, model_config, batch_infer_size=1, batch_timeout=10):
        super(TModelBase, self).__init__()
        self.inference_queue = inference_queue
        self.result_dict = result_dict
        self.batch_infer_size = batch_infer_size
        self.batch_timeout_in_sec = self._microsec_to_sec(batch_timeout)
        self.batch_timeout = batch_timeout
        self.model = self.model_init(model_config)

    def _microsec_to_sec(self, microsec):
        return microsec/1000

    def model_init(self, model_config: dict):
        raise NotImplementedError

    def predict(self, list_input):
        raise NotImplementedError

    # def run(self):
    #     list_inference = []
    #     list_request_id = []
    #     while True:
    #         t = current_milli_time()
    #         while (len(list_inference) < self.batch_infer_size) and current_milli_time() - t < self.batch_timeout:
    #             try:
    #                 [request_id, inp] = self.inference_queue.get(block=True, timeout=self.batch_timeout_in_sec)
    #                 list_inference.append(inp)
    #                 list_request_id.append(request_id)
    #             except Empty:
    #                 pass
    #             t = current_milli_time()
    #         list_result = self.predict(list_inference)
    #         for [_id, res] in zip(list_request_id, list_result):
    #             self.result_dict.update({_id:res})
    #         list_request_id.clear()
    #         list_inference.clear()

    def run(self):
        list_inference = []
        list_request_id = []
        while True:
            while True:
                try:
                    # , timeout=self.batch_timeout_in_sec)
                    [request_id, inp] = self.inference_queue.get(block=False)
                    list_inference.append(inp)
                    list_request_id.append(request_id)
                    if (len(list_inference) < self.batch_infer_size):
                        continue
                except Empty:
                    break
            if len(list_inference) != 0:
                list_result = self.predict(list_inference)
                for [_id, res] in zip(list_request_id, list_result):
                    self.result_dict.update({_id: res})
                list_request_id.clear()
                list_inference.clear()


class THandlerBase():
    def __init__(self, inference_queue, result_dict):
        self.inference_queue = inference_queue
        self.result_dict = result_dict

    def send_to_model(self, input_val):
        request_id = current_nano_time()
        input_ = self.process_input(input_val)
        self.inference_queue.put([request_id, input_], block=False)
        done = False
        while not done:
            result = self.result_dict.get(request_id, None)
            if result is not None:
                done = True
        return result

    def process_input(self, input):
        raise NotImplementedError


class ConnectionSink(threading.Thread):
    """Worker is a small helper to process incoming connection."""

    def __init__(self, thread_id, queue, processor_cls, handler_cls, inference_queue, result_dict):
        threading.Thread.__init__(self)
        self.thread_id = thread_id
        self.queue = queue
        self.processor_cls = processor_cls
        self.handler_cls = handler_cls
        self.inference_queue = inference_queue
        self.result_dict = result_dict

    def run(self):
        """Process queries from task queue, stop if processor is None."""
        self.handler = self.handler_cls(
            inference_queue=self.inference_queue, result_dict=self.result_dict)
        self.processor = self.processor_cls(self.handler)
        while True:
            try:
                iprot, oprot, otrans, callback = self.queue.get()
                # if iprot is None:
                #     continue
                self.processor.process(iprot, oprot)
                callback(True, otrans.getvalue())
            except Exception:
                logger.exception(
                    "Exception while processing request")
                callback(False, b'')


WAIT_LEN = 0
WAIT_MESSAGE = 1
WAIT_PROCESS = 2
SEND_ANSWER = 3
CLOSED = 4


def locked(func):
    """Decorator which locks self.lock."""

    def nested(self, *args, **kwargs):
        self.lock.acquire()
        try:
            return func(self, *args, **kwargs)
        finally:
            self.lock.release()
    return nested


def socket_exception(func):
    """Decorator close object on socket.error."""

    def read(self, *args, **kwargs):
        try:
            return func(self, *args, **kwargs)
        except socket.error:
            logger.debug('ignoring socket exception', exc_info=True)
            self.close()
    return read


class Message(object):
    def __init__(self, offset, len_, header):
        self.offset = offset
        self.len = len_
        self.buffer = None
        self.is_header = header

    @property
    def end(self):
        return self.offset + self.len


class Connection(object):
    """Basic class is represented connection.

    It can be in state:
        WAIT_LEN --- connection is reading request len.
        WAIT_MESSAGE --- connection is reading request.
        WAIT_PROCESS --- connection has just read whole request and
                         waits for call ready routine.
        SEND_ANSWER --- connection is sending answer string (including length
                        of answer).
        CLOSED --- socket was closed and connection should be deleted.
    """

    def __init__(self, new_socket, wake_up):
        self.socket = new_socket
        self.socket.setblocking(False)
        self.status = WAIT_LEN
        self.len = 0
        self.received = deque()
        self._reading = Message(0, 4, True)
        self._rbuf = b''
        self._wbuf = b''
        self.lock = threading.Lock()
        self.wake_up = wake_up
        self.remaining = False

    @socket_exception
    def read(self):
        """Reads data from stream and switch state."""
        assert self.status in (WAIT_LEN, WAIT_MESSAGE)
        assert not self.received
        buf_size = 8192
        first = True
        done = False
        while not done:
            read = self.socket.recv(buf_size)
            rlen = len(read)
            done = rlen < buf_size
            self._rbuf += read
            if first and rlen == 0:
                if self.status != WAIT_LEN or self._rbuf:
                    logger.error('could not read frame from socket')
                else:
                    logger.debug(
                        'read zero length. client might have disconnected')
                self.close()
            while len(self._rbuf) >= self._reading.end:
                if self._reading.is_header:
                    mlen, = struct.unpack('!i', self._rbuf[:4])
                    self._reading = Message(self._reading.end, mlen, False)
                    self.status = WAIT_MESSAGE
                else:
                    self._reading.buffer = self._rbuf
                    self.received.append(self._reading)
                    self._rbuf = self._rbuf[self._reading.end:]
                    self._reading = Message(0, 4, True)
                    done = True
            first = False
            if self.received:
                self.status = WAIT_PROCESS
                break
        self.remaining = not done

    @socket_exception
    def write(self):
        """Writes data from socket and switch state."""
        assert self.status == SEND_ANSWER
        sent = self.socket.send(self._wbuf)
        if sent == len(self._wbuf):
            self.status = WAIT_LEN
            self._wbuf = b''
            self.len = 0
        else:
            self._wbuf = self._wbuf[sent:]

    @locked
    def ready(self, all_ok, message):
        """Callback function for switching state and waking up main thread.

        This function is the only function witch can be called asynchronous.

        The ready can switch Connection to three states:
            WAIT_LEN if request was oneway.
            SEND_ANSWER if request was processed in normal way.
            CLOSED if request throws unexpected exception.

        The one wakes up main thread.
        """
        assert self.status == WAIT_PROCESS
        if not all_ok:
            self.close()
            self.wake_up()
            return
        self.len = 0
        if len(message) == 0:
            # it was a oneway request, do not write answer
            self._wbuf = b''
            self.status = WAIT_LEN
        else:
            self._wbuf = struct.pack('!i', len(message)) + message
            self.status = SEND_ANSWER
        self.wake_up()

    @locked
    def is_writeable(self):
        """Return True if connection should be added to write list of select"""
        return self.status == SEND_ANSWER

    # it's not necessary, but...
    @locked
    def is_readable(self):
        """Return True if connection should be added to read list of select"""
        return self.status in (WAIT_LEN, WAIT_MESSAGE)

    @locked
    def is_closed(self):
        """Returns True if connection is closed."""
        return self.status == CLOSED

    def fileno(self):
        """Returns the file descriptor of the associated socket."""
        return self.socket.fileno()

    def close(self):
        """Closes connection"""
        self.status = CLOSED
        self.socket.close()


class TMultiPoolServer(object):
    """TMultiPoolServer."""

    def __init__(self,
                 handler_cls,
                 processor_cls,
                 model_cls,
                 lsocket,
                 list_model_config,
                 inputProtocolFactory=None,
                 outputProtocolFactory=None,
                 batch_infer_size=1,
                 batch_timeout=10,
                 n_models=1,
                 n_handlers=10):
        self.model_cls = model_cls
        self.handler_cls = handler_cls
        self.processor_cls = processor_cls
        self.socket = lsocket
        self.in_protocol = inputProtocolFactory or TBinaryProtocolFactory()
        self.out_protocol = outputProtocolFactory or self.in_protocol
        self.n_handlers = int(n_handlers)
        self.n_models = int(n_models)
        self.clients = {}
        self.connection_queue = queue.Queue()
        # self.inference_queue = queue.Queue()
        self.inference_queue = multiprocessing.Queue()
        self.result_dict = multiprocessing.Manager().dict()
        # self.result_dict = dict()
        self.list_model_config = list_model_config
        self.batch_infer_size = batch_infer_size
        self.batch_timeout = batch_timeout

        self._read, self._write = socket.socketpair()
        self.prepared = False
        self._stop = False
        self.list_handlers = []
        self.list_models = []

    def set_num_handler(self, num):
        """Set the number of worker threads that should be created."""
        assert not self.prepared, "Can't change number of threads after start"
        self.n_handlers = num

    def prepare(self):
        """Prepares server for serve requests."""
        if self.prepared:
            return
        self.socket.listen()

        for model_config in self.list_model_config:
            process = self.model_cls(inference_queue=self.inference_queue,
                                     result_dict=self.result_dict,
                                     model_config=model_config,
                                     batch_infer_size=self.batch_infer_size,
                                     batch_timeout=self.batch_timeout)
            process.start()
            self.list_models.append(process)

        for idx in range(self.n_handlers):
            thread = ConnectionSink(idx, self.connection_queue,
                                    self.processor_cls, self.handler_cls,
                                    self.inference_queue, self.result_dict)
            thread.setDaemon(True)
            thread.start()
            self.list_handlers.append(thread)
        self.prepared = True

    def wake_up(self):
        """Wake up main thread.

        The server usually waits in select call in we should terminate one.
        The simplest way is using socketpair.

        Select always wait to read from the first socket of socketpair.

        In this case, we can just write anything to the second socket from
        socketpair.
        """
        self._write.send(b'1')

    def stop(self):
        """Stop the server.

        This method causes the serve() method to return.  stop() may be invoked
        from within your handler, or from another thread.

        After stop() is called, serve() will return but the server will still
        be listening on the socket.  serve() may then be called again to resume
        processing requests.  Alternatively, close() may be called after
        serve() returns to close the server socket and shutdown all worker
        threads.
        """
        self._stop = True
        self.wake_up()

    def _select(self):
        """Does select on open connections."""
        readable = [self.socket.handle.fileno(), self._read.fileno()]
        writable = []
        remaining = []
        for i, connection in list(self.clients.items()):
            if connection.is_readable():
                readable.append(connection.fileno())
                if connection.remaining or connection.received:
                    remaining.append(connection.fileno())
            if connection.is_writeable():
                writable.append(connection.fileno())
            if connection.is_closed():
                del self.clients[i]
        if remaining:
            return remaining, [], [], False
        else:
            return select.select(readable, writable, readable) + (True,)

    def handle(self):
        """Handle requests.

        WARNING! You must call prepare() BEFORE calling handle()
        """
        assert self.prepared, "You have to call prepare before handle"
        rset, wset, xset, selected = self._select()
        for readable in rset:
            if readable == self._read.fileno():
                # don't care i just need to clean readable flag
                self._read.recv(1024)
            elif readable == self.socket.handle.fileno():
                try:
                    client = self.socket.accept()
                    if client:
                        self.clients[client.handle.fileno()] = Connection(client.handle,
                                                                          self.wake_up)
                except socket.error:
                    logger.debug('error while accepting', exc_info=True)
            else:
                connection = self.clients[readable]
                if selected:
                    connection.read()
                if connection.received:
                    connection.status = WAIT_PROCESS
                    msg = connection.received.popleft()
                    itransport = TTransport.TMemoryBuffer(
                        msg.buffer, msg.offset)
                    otransport = TTransport.TMemoryBuffer()
                    iprot = self.in_protocol.getProtocol(itransport)
                    oprot = self.out_protocol.getProtocol(otransport)
                    self.connection_queue.put([iprot, oprot,
                                               otransport, connection.ready])
        for writeable in wset:
            self.clients[writeable].write()
        for oob in xset:
            self.clients[oob].close()
            del self.clients[oob]

    def close(self):
        """Closes the server."""
        for _ in range(self.n_handlers):
            self.connection_queue.put([None, None, None, None, None])
        self.socket.close()
        self.prepared = False

    def serve(self):
        """Serve requests.

        Serve requests forever, or until stop() is called.
        """
        self._stop = False
        self.prepare()
        while not self._stop:
            self.handle()
